#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <time.h>
#include <sys/resource.h>
#include <sys/wait.h>     
#include <math.h>

#include "time_util.h"
#include "tsc_lib.h"

#define WAIT 999
#define MILLION 1000000
#define FALSE 0
#define TRUE  1

void init();
void fork_exec(char *filename, int iter, int warmup=0);
void sigchild_handler(int n);

int main(int argc, char **argv)
{
  int iterations, warmup_iter = 0;

  if ((argc != 3) && (argc != 4)){
    fprintf(stderr, "Usage: %s <filename> <iterations> [warmup iterations]\n",
	    argv[0]);
    exit(1);
  }
  
  init();
  iterations = atoi(argv[2]);
  if (argc == 4)
    warmup_iter = atoi(argv[3]);

  if (warmup_iter)
    fork_exec(argv[1], warmup_iter, TRUE);
  
  fork_exec(argv[1], iterations, FALSE);

  return 0;
}
 
void init()
{
  struct sigaction act;                /* signal setting structure */
  act.sa_handler = sigchild_handler;   /* the handler (function) */
  act.sa_flags   = SA_NOCLDSTOP;       /* SIGSTOP does not cause SIGCHLD */
  sigemptyset(&act.sa_mask);
  
  if (sigaction(SIGCHLD , &act , NULL) != 0) {
    fprintf(stderr, "sigaction for SIGCHLD failed");
    perror("init");
    exit(1);
  }
  
  tsc_init(500); //500 MHz machine
  return;
}

void fork_exec(char *filename, int iter, int warmup)
{
  int i, pid;
  static char *arg_list[10];
  tsc_counter_t t_before, t_after;
  double  t_dif, std_t, square_t, sum_t;

  arg_list[0] = filename;
  arg_list[1] = NULL;
  
  square_t =  sum_t = 0;
  for (i=0; i<iter; i++){   
    t_before = tsc_readCycles();
    pid = fork();
    if (pid < 0) {
      fprintf(stderr, "fork_exec: failed to fork");
      perror("fork_exec");
      exit(1);
    } else if (pid == 0) {
      if (execvp(filename, arg_list) < 0)
	perror("fork_exec");
      else 
	exit(0); /* just to make sure */
    }
    sigchild_handler(WAIT);
    t_after = tsc_readCycles();
    t_dif   = ((double)(tsc_counterToNsec(t_after) -  tsc_counterToNsec(t_before))) / 1000.0;
    square_t += t_dif*t_dif;
    sum_t    += t_dif;
  }
  if (warmup) /* no need for calculations when warming up */
    return;
  
  std_t = sqrt((square_t /(double)iter) - (sum_t /(double)iter)*(sum_t /(double)iter));   
   
  fprintf(stderr, "Avg. time (Milli) %5.3f\n", (sum_t/(double)iter)/1000.0);
  fprintf(stderr, "StdDev (Milli)    %5.3f\n", std_t/1000.0);
}

void sigchild_handler(int n)
{
  pid_t          wpid, pid;
  int		 status;
  struct rusage  rusage;
  int wait_options;

  if (n == WAIT)
    wait_options = 0;
  else
    wait_options = WNOHANG;

  while (1) {
    /*
     * call wait4 to get the status of a terminated sprocess
     */
    wpid = -1;
    if ((pid = wait4(wpid, &status, wait_options, &rusage)) == -1) {
      return;
    }
    if (pid == 0) {
      return;
    }
  }
}

