/*
  Time Stamp Counters library. Written by Yoav Etsion, January 2000.
  
  This library offers a tool for measuring time without the overhead of a
  system call. It uses a Pentium op-code that reads the Pentium's clock
  cycles counter, at user level.
  When compiling an application that uses this library, make sure you compile
  with an optimization flag that inlines functions:
  GNU: -O and up.
  WIN: /O2
  QNX: -Oe

  Note: All time and cycles measurements are taken since the CPU was last
  reset.

  Data Types:
	The library provides two data structures and the functions to
	manipulate them:
  1. tsc_counter_t - the storage type of the clock cycle counter. Due to
  efficiency problems on platforms that do not support a 64-bit integral
  type (QNX with watcom), we found it necessary to avoid any translations to
  integral types at the measurement time (it will take too long), but rather
  keep the raw data in an internal data structure, and process it later.
  2. tsc_counter_int_t - the integral type, large enough to hold a 64-bit
  unsigned integral value, which is how the counter is kept inside the CPU
  itself. We give the user an abstraction of the actual type, because not
  all platforms support such integral types, and sometimes we have to emulate
  using double values. The only assurance the library gives a user is that
  this type can be used with all mathemtical operators (+,-,*,/ etc.) to
  calculate time differences. The library also provides the user with a
  function that translate the integral value into a character array.

  Although seperating the measurment type from the calculation type, we gave
  the user the option the take measurements directly using the integral type,
  but we offer no efficiency guarentees and highly recommend against it (as
  mentioned before, these efficiency problems arise only on platforms that do
  not support a 64-bit integral type, so there are no such problems on unix
  with a GNU compilers, or with VC).

  Functions API:

  1. 	
	int tsc_init();

	  Initializes the library's internal data. This function must be
	  called only once, and	before any other library function in order
	  for the data to be consistent.
	Note: this function might take up to 1 second to run, for an accurate
	cpu measurement.
  

  2.
	INLINE tsc_counter_int_t tsc_getIntegral(tsc_counter_t counter)

	Converts the counter_t type, which holds two 32 bit integral type, to
	a primitive integral type, that has all the math operators declared
	for.

  3.
	INLINE tsc_counter_t tsc_readCycles() 
  
	  The main issue. Takes a reading and returns the counter. We do not
	  return a counter_int_t to avoid any unnecessary calculation during
	  a measurement. The calculation should follow the measurement rather
	  than influence it. If a user wants the measurement done in a
	  primitive type, he can use the following function. A user that does
	  so must take into account that the measurement itself will take
	  longer (although still a factor faster and more accurate
	  than any system call).
  
  4.
	INLINE tsc_counter_int_t tsc_readIntCycles() 
  
	  Same as the last function. Returns the primitive type, but takes
	  longer on compilers that do not support a 64-bit integral type.
	  
  5.
         const char* tsc_integralToStr(tsc_counter_int_t counter,
	                               unsigned int char_num);
	 const char* tsc_counterToStr(tsc_counter_t counter,
	                              unsigned int char_num);
  
	Return a value as a string. Used to print the value.
	Parameters:
	counter: need I say more?
	char_num: the number of characters to print. If the actual value is
	bigger, only prints the least significant digits. If char_num is 0
	the string is not truncated.
	Note: returns a pointer into a static buffer. Value will change after
	another call.
	
  6.
        INLINE unsigned long tsc_CPUSpeed() 

	Returns the CPU speed calculated in the init function. The CPU speed
	is measured and then compared to the possible list of values. This
	list can be changed by editing the TSC_CPU_SPEEDS array, declared
	later in this file. The value picked is the closest value among those
	which are within a certain proximity to the measured value. The
	proximity factor can be changed by editing the TSC_CPU_PERCENT const
	declared later in this file (the value is the percentage).

  7.
	tsc_counter_int_t tsc_counterToNsec(tsc_counter_t counter);
	tsc_counter_int_t tsc_integralToNsec(tsc_counter_int_t counter);
  
	Translate the clock cycles counter into a more human form of
	time: nanoseconds.
	Note: This will lose some accuracy for a processor faster than 1GHz,
	because each cycle will then take less than a nanosecond (read the
	limitations clause).

  Limitations: 

  Because we are using some hacks to maintain a 64-bit integral type, certain
  counter might overflow in specific situations:
  I.   If the computer is up for a period of exactly 185 days, the nanosecond
  counter will overflow. Just make sure that any strange measurements you get
  are not caused by this effect.
  II.  The library is not SMP safe! To make it so we must lock the process
  in a CPU, so we won't	get different values when migrating from one CPU t
  another.
  III. The most strict but least significant limit: The CPU's Time stamp
  Counter will, of course, overflow after 2^64 cycles (64-bit counter). For a
  1GHz processor this will occure after:
	2^64 cycles =~ 2^34 seconds (2^30 = 1G) =~ 198841 days =~ 544 Years,
	so we realy don't think it matters...

  Platforms:
  The library was compiled and tested on the following platforms:
  I.   Linux and BSD (using GNU gcc and g++)
  II.  Windows NT (using Visual C++)
  III. QNX RTOS (using the annoying Watcom C compiler. Make sure you compile
  with -Oe).
  In order to port the library to a new platform, just write a file mapping
  the platform indepentent interface documented later in this file
  (and implemented in tsc_lib_GNUC.h, tsc_lib_QNX.h and tsc_lib_WIN32.h -
  each file implementd the interface for the implied platform). Next, one
  should add his new file an include protected by appropreate #ifdef
  for the new platform.
*/

#ifndef TSC_LIB_HH
#define TSC_LIB_HH

#define TSC_OK    1
#define TSC_ERROR 0

#ifdef __cplusplus
extern "C" {
#endif

/*
  This is our way of accessing each half of the 64-bit integral type.
  Also, we need it in the platform definition files.
*/
typedef struct {
  unsigned long low;
  unsigned long high;
} tsc_int64;

/*
  Since we are dealing with some non-standard features of the compiler here,
  we must resort to some platform specific definitions. Each supported
  platform has a definitions h-file and we simply include the appropeate one.
  These non-standard features are:
 
  1. A 64-bit integral type. Some platforms don't support such a type, and
  those who do, don't maintain a standard name but have some very creative
  ones (long long, int64 etc.). That is why we have to map it into 'our' type
  name on every system that has such a type, and we must also define a compile
  time flag to indicate if the platform supports such a type.
  
  2. A function to print the 64-bit integral type into a string. Because
  this type is not standard, every platform has it's method of printing this
  giant.

  3. Inline function directive. Again, a simple mapping problem: some systems
  call it inline, others __inline__ and sometimes __inline. We simply map our
  directive called INLINE to the correct one on the current platform.
 
  4. Inline assembly syntax: We need to define a way to access the Time Stamp
  Counter on the CPU, using assembly. We define a macro that performs this
  action for each platform.
  
  5. In order to measure the CPU speed we need a time measurement, and a sleep
  function. Each platform has different system calls for these purposes, so we
  have our own inline functions (library private ones), that get the time and
  sleep with milisecond resolution, assuming most platform support such low
  standards nowadays.
  */
#if defined(__GNUC__)  && (defined(__linux__) || defined(__bsdi__))
/* BSD and Linux with GNU's compilers use this definition file. */
#include "tsc_lib_GNUC.h"

#elif defined(_WIN32)
/* This is for windows with visual C++ */
#include "tsc_lib_WIN32.h"

#elif defined(__QNX__) && defined(__WATCOMC__)
/* QNX with Watcom's compiler. */
#include "tsc_lib_QNX.h"

#else

#error "Your platform/compiler combination is not supported. Please check the documentation"

#endif

/*
  The basic counter type. Due to efficiency, we save the value just
  as it is received - two longs. If the compiler support a 64 bit 
  integral type, the conversion if free using union. Otherwise, we define
  a macro to convert the value into a double. 
  Note: The user must never use this type directly for any calculations, but
  rather use the counter_int_t. Both type do not necessarily share the same
  init time, i.e. they
  will show different times.
*/
typedef union {
#if defined(HAVE_ULONGLONG)
  TSC_ULONGLONG full;
#endif
  tsc_int64 bit;
} tsc_counter_t;

/*
  The type used for any calculations.
  We define a counter primitive type, that is big enough to store the 64-bit
  counter. This type promised the user that all the mathematical and logical
  operators can be used for it.	Unfotunately, not all systems support a native
  64-bit integral type. If one is not present, we simulate it using a double
  precision type.
*/
typedef TSC_ULONGLONG tsc_counter_int_t;

/*
  Initializes the library's internal data. This function must be called only
  once, and before any other library function in order for the data to be
  consistent.
  Note: this function might take up to 1 second to run, for an accurate cpu
  measurement.
*/
int tsc_init();

/*
  Converts the counter_t type, which holds two 32 bit integral type, to a
  primitive type, that has all the math operators declared for.
  */
INLINE tsc_counter_int_t tsc_getIntegral(tsc_counter_t counter)
{
/*
  If we have counter_int_t is a native 64-bit integer, we can simply return
  the 'full' field of the union. Otherwise, we have to compute the value from
  the two unsigned longs.
*/
#ifdef HAVE_ULONGLONG
  return counter.full;
#else
  /* 2^32 as a double. Needed for the 2 longs
     to counter conversion. */
#define exp32d (4294967296.0)
  
  return (double)counter.bit.high*exp32d + counter.bit.low;
#endif
}

/*
  The main issue. Takes a reading and returns the counter. We do not return a
  counter_int_t to avoid any unnecessary calculation during a measurement.
  The calculation should follow the measurement rather than influence it. If
  a user want the measurement done in a primitive type, he can use the
  following function. A user that does so must take into account that the
  measurement itself will take longer (although still a factor faster and
  more accurate than any system call).
  */
INLINE tsc_counter_t tsc_readCycles() 
{
  tsc_counter_t c;
  rdtsc(c.bit);

  return c;
}

/*
  Same as the last function, only returns the primitive type, but takes longer.
*/
INLINE tsc_counter_int_t tsc_readIntCycles() 
{
  tsc_counter_t c;
  rdtsc(c.bit);

  return tsc_getIntegral(c);
}

/*
  Return a value as a string. Used to print the value.
  Parameters:
  counter: need I say more?
  char_num: the number of characters to print. If the actual value is bigger,
  only prints the least significant digits. If char_num is 0 the string is
  not truncated.
  Note: returns a pointer into a static buffer. Value will change after
  another call.
*/
const char* tsc_integralToStr(tsc_counter_int_t counter, unsigned int char_num);
const char* tsc_counterToStr(tsc_counter_t counter, unsigned int char_num);
/*
  The two consts relate to the function:
  TSC_CPU_SPEEDS: The possible speeds this library recognizes.
  TSC_CPU_PERCENT: The maximum diviation bewteen the estimated speed and
  the closest possible speed we allow.
  */
extern unsigned long g_cpuSpeed;
static const unsigned long TSC_CPU_SPEEDS[] = { 
    90, 120, 133, 166, 200, 233, 266, 300, 333, 350, 366, 400, 433, 450, 466, 
    500, 533, 550, 566, 600, 633, 650, 666, 700, 733, 750, 766, 800, 833, 850,
    866, 900, 1000
};
static const double TSC_CPU_PERCENT = 0.05;
INLINE unsigned long tsc_CPUSpeed() 
{
	return g_cpuSpeed;
}

/*
  Translate the clock cycles counter into a more human form of time:
  nanoseconds.
  Note: This will lose some accuracy for a processor faster than 1GHz,
  because each cycle will then take less than a second.
*/
tsc_counter_int_t tsc_counterToNsec(tsc_counter_t counter);
tsc_counter_int_t tsc_integralToNsec(tsc_counter_int_t counter);


/*
  TBD: I should add a tsc_init and tsc_close which changes the tsc bit on 
  the cr4 register to the correct value, but this requires writing a 
  basic driver for each operating system (BSD, Linux, NT, QNX). Since 
  these OSs set the tsc bit to allow user level queries, it is not urgent.
  */

#ifdef __cplusplus
}
#endif

#endif  /* TSC_LIB_HH */

