/******************************************************************
This program is used to check the time it takes to create a process
in Win NT.
Run it from the command line with the following arguments:
<filename> <iterations> [warmup iterations]
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <math.h>
//#include "timeutil.h"


// if set, wait for the child proc to die before spawning a new one
#define WAIT_FOR_CHILDREN 1 
#define MAX_CHILDREN      10000
#define MAX_DATAPOINTS 50000
#define SLICE_FRAC .1
#define TRUE 1
#define FALSE 0
#define MILLION 1000000

int g_iterations_per_slice;
/* Used to measure the average time of each slice */
LARGE_INTEGER g_slice_datapoints[MAX_DATAPOINTS];

FILETIME CreationTime, ExitTime, KernelTimeBefore, UserTimeBefore, 
		 KernelTimeAfter, UserTimeAfter;
FILETIME ChildrenUserTime[MAX_CHILDREN], ChildrenKernelTime[MAX_CHILDREN],
         TotalChildrenUserTime, TotalChildrenKernelTime;


void CalcTimes(int iter);
void CalcChildrenTimes(int iter);
void CalcStdev(int datapoints);
double TimeDif(LARGE_INTEGER TimeAfter, LARGE_INTEGER TimeBefore);
void ProcRun(char *path, int iter, int warmup_iter = 0);

SYSTEMTIME SubtractFileTimes(const FILETIME *after, const FILETIME *before);

void main(int argc, char **argv)
{
  int iter, warmup_iter = 0;

  if ((argc != 3) && (argc !=4)){
    fprintf(stderr, "Usage: %s <app filename> <iterations> [warmup iterations]\n",
            argv[0]);
    exit(1);
  }
   
  iter = atoi(argv[2]);
  if (argc == 4)
	warmup_iter = atoi(argv[3]);

  if (warmup_iter != 0)
	  ProcRun(argv[1], warmup_iter, true);
  ProcRun(argv[1], iter);
}

void ProcRun(char *path, int iter, BOOL warmup)
{
  int datapoints_count = 0;
  LARGE_INTEGER TimeBefore = {0,0}, TimeAfter = {0,0}, freq = {0,0};
  STARTUPINFO si; 
  PROCESS_INFORMATION info;
  int ret_val;
  
 
  HANDLE hProcess = GetCurrentProcess();
  if (iter == 0) {
	fprintf(stderr, "Error: number if iterations is 0.\n");
	exit(1);
  }

  fprintf(stderr, "Running file %s, %d times\n", path, iter);


  // Set up members of STARTUPINFO structure. 
  ZeroMemory( &si, sizeof(STARTUPINFO) );
  si.cb = sizeof(STARTUPINFO); 
  si.lpDesktop = "WinSta0\\Default"; 
  si.dwFlags = STARTF_USESTDHANDLES;

  /*if (GetProcessTimes(
		 hProcess,           
		 &CreationTime,       
		 &ExitTime,           
		 &KernelTimeBefore,   
         &UserTimeBefore) == 0) {
		fprintf(stderr, "GetProcessTimes failed.\n");
		exit(1);
	}*/
  g_iterations_per_slice = (int)(iter *SLICE_FRAC + 0.5);
  if (g_iterations_per_slice == 0) {
    g_iterations_per_slice = 1000; /* dummy value for the warmup iterations */
  }
  QueryPerformanceCounter(&TimeBefore);
  for (int i=0; i<iter; i++) {
	if ((i % g_iterations_per_slice) == 0){
	  QueryPerformanceCounter(&g_slice_datapoints[datapoints_count]);
      datapoints_count++;
    }
	ret_val = CreateProcess(
		path, 
		"",          // command line 
		NULL,          // process security attributes 
		NULL,          // primary thread security attributes
		TRUE,          // handles are inherited 
		DETACHED_PROCESS,             // creation flags
		NULL,          // use parent's environment 
		NULL,          // use parent's current directory 
		&si,		     // STARTUPINFO pointer 
		&info);  // receives PROCESS_INFORMATION 

	if (! ret_val) {
		fprintf(stderr, "Error: CreateProcess failed %d\n", GetLastError());
		exit(1);
	}
	
	if (WAIT_FOR_CHILDREN) {
		if (WaitForSingleObject (info.hProcess, INFINITE) == WAIT_FAILED){
			fprintf(stderr, "WaitForSingleObject failed :%d", GetLastError());
		}
	}

	// just collect the kernel of user time of the child, sum it later
	/*if (GetProcessTimes(
		 info.hProcess,           
		 &CreationTime,       
		 &ExitTime,           
		 &ChildrenKernelTime[i],   
         &ChildrenUserTime[i]) == 0) {
			fprintf(stderr, "GetProcessTimes failed.\n");
			exit(1);
		}*/
	
	CloseHandle(info.hProcess);
	CloseHandle(info.hThread);
  }
  QueryPerformanceCounter(&TimeAfter);
  QueryPerformanceCounter(&g_slice_datapoints[datapoints_count]);
  datapoints_count++;

  /*
  if (GetProcessTimes(
		 hProcess,
		 &CreationTime,       
		 &ExitTime,           
		 &KernelTimeAfter,   
         &UserTimeAfter) == 0) {
			fprintf(stderr, "GetProcessTimes failed.\n");
			exit(1);
		}*/
  if (warmup)
	  return;

  double dif = TimeDif(TimeAfter, TimeBefore);
  fprintf(stderr, "Average time to Create a process is %3.3f MicroSecond\n", dif/iter);
  CalcStdev(datapoints_count);
	
  //CalcChildrenTimes(iter);	
}


double TimeDif(LARGE_INTEGER TimeAfter, LARGE_INTEGER TimeBefore)
{
  if (TimeAfter.HighPart != TimeBefore.HighPart) {
	fprintf(stderr, "Internal Error\n");
	exit(1);
  }
  double dif = TimeAfter.LowPart - TimeBefore.LowPart;
  
  /* query for the counter frequency */
  LARGE_INTEGER freq;
  if (QueryPerformanceFrequency(&freq) == 0) {
	fprintf(stderr, "QueryPerformanceFrequency not supported.\n");
	exit(1);
  }
  dif *= (1/((float)freq.LowPart))*MILLION;	

  return dif;
}

void CalcStdev(int datapoints)
{  
  static double slice_avg[MAX_DATAPOINTS];
  double diff, slices_avg, sum, stdev;
  double slice_total_time;  
  int i;
  double slices_avg_sum = 0.;

  for (i=1; i<datapoints; i++){
    slice_total_time = TimeDif(g_slice_datapoints[i], g_slice_datapoints[i-1]);
    slice_avg[i-1] =  slice_total_time / ((float)g_iterations_per_slice);
    slices_avg_sum += slice_avg[i-1];
  }

  slices_avg = slices_avg_sum / ((datapoints-1) * 1.0);
  sum = 0;
  for (i=0; i<datapoints-1; i++){
    diff = slice_avg[i] - slices_avg;
    sum += diff*diff;
  }
  stdev = sqrt(sum / (datapoints - 1));
  stdev *= sqrt(g_iterations_per_slice);
  
  fprintf(stderr, "The STDDEV of %d slices is: %3.3f\n", datapoints-1, stdev);
}

/*
void old()
{
SYSTEMTIME before_kerneltime, before_usertime;
  SYSTEMTIME after_kerneltime, after_usertime;

  if (FileTimeToSystemTime (&KernelTimeBefore, &before_kerneltime) == 0){
	fprintf(stderr, "FileTimeToSystemTime failed.\n");
	exit(1);
  }
  if (FileTimeToSystemTime (&UserTimeBefore, &before_usertime) == 0){
	fprintf(stderr, "FileTimeToSystemTime failed.\n");
	exit(1);
  }

  if (FileTimeToSystemTime (&KernelTimeAfter, &after_kerneltime) == 0){
	fprintf(stderr, "FileTimeToSystemTime failed.\n");
	exit(1);
  }
  if (FileTimeToSystemTime (&UserTimeAfter, &after_usertime) == 0){
	fprintf(stderr, "FileTimeToSystemTime failed.\n");
	exit(1);
  }

  SYSTEMTIME dif;
  dif = SubtractFileTimes(&KernelTimeAfter, &KernelTimeBefore);
  fprintf(stderr, "Total kernel time: %d:%d (seconds:miliseconds)\n",
			dif.wSecond, dif.wMilliseconds);
  dif = SubtractFileTimes(&UserTimeAfter, &UserTimeBefore);
  fprintf(stderr, "Total user time: %d:%d (seconds:miliseconds)\n", 
				dif.wSecond, dif.wMilliseconds);
}*/

/*
SYSTEMTIME SubtractFileTimes(const FILETIME *after, const FILETIME *before)
{
	SYSTEMTIME result_systemtime;
		
	LARGE_INTEGER time_before, time_after, time_dif;
	FILETIME result_filetime;
	time_before.HighPart = before->dwHighDateTime;
	time_before.LowPart  = before->dwLowDateTime;
	time_after.HighPart  = after->dwHighDateTime;
	time_after.LowPart   = after->dwLowDateTime;

	time_dif.QuadPart = time_after.QuadPart-time_before.QuadPart;
	result_filetime.dwHighDateTime = time_dif.HighPart;
	result_filetime.dwLowDateTime = time_dif.LowPart;

	if (FileTimeToSystemTime (&result_filetime, &result_systemtime) == 0){
		fprintf(stderr, "FileTimeToSystemTime failed.\n");
		exit(1);
	}

	return result_systemtime;
}*/

/*
 * Sum the children user and kernel time and print the sum and the average
 *
void CalcChildrenTimes(int iter)
{
	SYSTEMTIME result;
	DWORD total_kernel_time = 0, total_user_time = 0;
	DWORD sec, mili;

	*
	LARGE_INTEGER total_kernel_time, total_user_time,
				  children_kernel_time, children_user_time;
	children_kernel_time.QuadPart = children_user_time.QuadPart = 
		total_kernel_time.QuadPart = total_user_time.QuadPart = 0;
	*

	for (int i=0; i<iter; i++){
		total_kernel_time += FileTime2Mili(ChildrenKernelTime[i]);
		total_user_time   +=   FileTime2Mili(ChildrenUserTime[i]);
	}



	 *
	  * Total kernel time and user time
	  *
	TraslateTimeToMili(total_kernel_time, sec, mili);
	fprintf(stderr, "Total children kernel time: %d:%d (seconds:miliseconds)\n",
			sec, mili);

	TraslateTimeToMili(total_user_time, sec, mili);
	fprintf(stderr, "Total children user time: %d:%d (seconds:miliseconds)\n",
			sec, mili);
	

	/*
	 * Average kernel time and user time
	 
	total_kernel_time.QuadPart /= iter;
	total_user_time.QuadPart /= iter;
	
	result = LargeInt2SystemTime(total_kernel_time);
	fprintf(stderr, "Average children kernel time: %d:%d (seconds:miliseconds)\n",
			result.wSecond, result.wMilliseconds);

	result = LargeInt2SystemTime(total_user_time);
	fprintf(stderr, "Average children user time: %d:%d (seconds:miliseconds)\n",
			result.wSecond, result.wMilliseconds);
			
}*/

