/******************************************************************
This program is used to check the time it takes to create a process
in Win NT.
Run it from the command line with the following arguments:
<filename> <iterations> [warmup iterations]
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <math.h>

#define MAX_DATAPOINTS 50000
#define SLICE_FRAC .1
#define TRUE 1
#define FALSE 0
#define MILLION 1000000

//function prototypes
void RunProc(char *filename, int instances);
void SuspendResume(int iter, BOOL warmup);
double TimeDif(LARGE_INTEGER TimeAfter, LARGE_INTEGER TimeBefore);
void CalcStdev(int datapoints);

void TraslateTimeToMili(DWORD Time, DWORD &Min, DWORD &Mili);
void TraslateTimeToMicro(DWORD Time, DWORD &Min, DWORD &Micro);
SYSTEMTIME SubtractFileTimes(const FILETIME *after, const FILETIME *before);
long CalcAvgFileTime(const SYSTEMTIME *time, int iter);

//global variables
HANDLE hProc0Thread, hProc1Thread, hProc0, hProc1;
int g_iterations_per_slice;
/* Used to measure the average time of each slice */
LARGE_INTEGER g_slice_datapoints[MAX_DATAPOINTS];

int main(int argc, char** argv)
{
  int warmup_iter = 0;
  int iterations;
 
  if ((argc != 3) && (argc != 4)){
    fprintf(stderr, "Usage: %s <suspend-resume iterations> [warmup iterations]\n", argv[0]);
    exit(1);
  }
  iterations = atoi(argv[2]);
  if (argc == 4)
    warmup_iter = atoi(argv[3]);

  RunProc(argv[1], 2);
  if (warmup_iter)
	SuspendResume(iterations, TRUE);
  SuspendResume(iterations, FALSE);
  TerminateProcess(hProc0, 0);
  TerminateProcess(hProc1, 0);
  return 0;
}


/*
 * Create a few instances of the given executable
 */
void RunProc(char *filename, int instances)
{
  STARTUPINFO si; 
  PROCESS_INFORMATION info;
  int ret_val;
   

  fprintf(stderr, "Running file %s\n", filename);
   // Set up members of STARTUPINFO structure. 
  ZeroMemory( &si, sizeof(STARTUPINFO) );
  si.cb = sizeof(STARTUPINFO); 
  si.lpDesktop = "WinSta0\\Default"; 
  si.dwFlags = STARTF_USESTDHANDLES;
  for (int i=0; i<instances; i++) {
	ret_val = CreateProcess(
		filename, 
		"",          // command line 
		NULL,          // process security attributes 
		NULL,          // primary thread security attributes
		TRUE,          // handles are inherited 
		DETACHED_PROCESS,             // creation flags
		NULL,          // use parent's environment 
		NULL,          // use parent's current directory 
		&si,		     // STARTUPINFO pointer 
		&info);  // receives PROCESS_INFORMATION 

	if (! ret_val) {
		fprintf(stderr, "Error: CreateProcess failed %d\n", GetLastError());
		exit(1);
	}

	//CloseHandle(info.hProcess);
	//CloseHandle(info.hThread);
	if (i == 0){
		hProc0Thread = info.hThread;
		hProc0 = info.hProcess;
	} else{
		hProc1Thread = info.hThread;
		hProc1 = info.hProcess;
	}
  }
}

void SuspendResume(int iter, BOOL warmup)
{
  int i, running_index;
  int datapoints_count = 0;
  LARGE_INTEGER TimeBefore = {0,0}, TimeAfter = {0,0}, freq = {0,0};

  SuspendThread(hProc0Thread);
  running_index = 0;
  if (warmup)
	fprintf(stderr, "warming up %d iterations\n", iter);
  else
	fprintf(stderr, "suspending & resuming %d iterations.\n", iter);
   g_iterations_per_slice = (int)(iter *SLICE_FRAC + 0.5);
  if (g_iterations_per_slice == 0) {
    g_iterations_per_slice = 1000; /* dummy value for the warmup iterations */
  }
  QueryPerformanceCounter(&TimeBefore);
  for (i=0; i<iter; i++) { 
	if ((i % g_iterations_per_slice) == 0){
	  QueryPerformanceCounter(&g_slice_datapoints[datapoints_count]);
      datapoints_count++;
    }
    if (running_index == 0){
		SuspendThread(hProc1Thread); //remember that there is a "suspend count".
		ResumeThread(hProc0Thread);
		running_index = 1;
    } else {
      SuspendThread(hProc0Thread);
      ResumeThread(hProc1Thread);
      running_index = 0;
    }
  }
  QueryPerformanceCounter(&TimeAfter);
  QueryPerformanceCounter(&g_slice_datapoints[datapoints_count]);
  datapoints_count++;

  if (warmup){ /* just warming up, no need for calculations */
    if (running_index == 1){
      SuspendThread(hProc0Thread);
      ResumeThread(hProc1Thread);; 
    }
    return;
  }
 
  double dif = TimeDif(TimeAfter, TimeBefore);
  fprintf(stderr, "Average time to suspend-resume is %3.3f MicroSecond\n", dif/iter);
  CalcStdev(datapoints_count);
}

double TimeDif(LARGE_INTEGER TimeAfter, LARGE_INTEGER TimeBefore)
{
  if (TimeAfter.HighPart != TimeBefore.HighPart) {
	fprintf(stderr, "Internal Error\n");
	exit(1);
  }
  double dif = TimeAfter.LowPart - TimeBefore.LowPart;
  
  /* query for the counter frequency */
  LARGE_INTEGER freq;
  if (QueryPerformanceFrequency(&freq) == 0) {
	fprintf(stderr, "QueryPerformanceFrequency not supported.\n");
	exit(1);
  }
  dif *= (1/((float)freq.LowPart))*MILLION;	

  return dif;
}

void CalcStdev(int datapoints)
{  
  static double slice_avg[MAX_DATAPOINTS];
  double diff, slices_avg, sum, stdev;
  double slice_total_time;  
  int i;
  double slices_avg_sum = 0.;

  for (i=1; i<datapoints; i++){
    slice_total_time = TimeDif(g_slice_datapoints[i], g_slice_datapoints[i-1]);
    slice_avg[i-1] =  slice_total_time / ((float)g_iterations_per_slice);
    slices_avg_sum += slice_avg[i-1];
  }

  slices_avg = slices_avg_sum / ((datapoints-1) * 1.0);
  sum = 0;
  for (i=0; i<datapoints-1; i++){
    diff = slice_avg[i] - slices_avg;
    sum += diff*diff;
  }
  stdev = sqrt(sum / (datapoints - 1));
  stdev *= sqrt(g_iterations_per_slice);
  
  fprintf(stderr, "The STDDEV of %d slices is: %3.3f\n", datapoints-1, stdev);
}



void TraslateTimeToMili(DWORD Time, DWORD &Sec, DWORD &Mili)
{
	Sec = Time / 1000;
	Mili = Time % 1000;
}

void TraslateTimeToMicro(DWORD Time, DWORD &Sec, DWORD &Micro)
{
	Sec = Time / 1000000;
	Micro = Time % 1000000;
}

/*
void old_calculations()
{
    DWORD TimeBefore, TimeAfter, TotalTime, TotalSec, TotalMili, AvgMin, AvgMicro;
  double AvgTime;
  FILETIME CreationTime, ExitTime, KernelTimeBefore, UserTimeBefore, 
		   KernelTimeAfter, UserTimeAfter;

   
	if (GetProcessTimes(
		 hProcess,           
		 &CreationTime,       
		 &ExitTime,           
		 &KernelTimeAfter,   
         &UserTimeAfter) == 0) {
		fprintf(stderr, "GetProcessTimes failed.\n");
		exit(1);
  }

  SYSTEMTIME TotalKernelTime, TotalUserTime;
  TotalKernelTime = SubtractFileTimes(&KernelTimeAfter, &KernelTimeBefore);
  TotalUserTime   = SubtractFileTimes(&UserTimeAfter,   &UserTimeBefore);	
  
  TotalTime = TimeAfter - TimeBefore;
  TraslateTimeToMili(TotalTime, TotalSec, TotalMili);
  AvgTime = (TotalTime * 1000)/ iter;
  //TraslateTimeToMicro(AvgTime, AvgMin, AvgMicro);	
  fprintf(stderr, "Average time %4.4f (MicroSeconds) ; Total time %d:%d (Sec:MiliSec)\n",
		  AvgTime, TotalSec, TotalMili);
  fprintf(stderr, "Average Kernel time %d; Total time %d:%d (Sec:MiliSec)\n", 
		  CalcAvgFileTime(&TotalKernelTime, iter),
		  TotalKernelTime.wSecond, 
		  TotalKernelTime.wMilliseconds);
  fprintf(stderr, "Average User time   %d; Total time %d:%d (Sec:MiliSec)\n", 
		CalcAvgFileTime(&TotalUserTime, iter), 
		TotalUserTime.wSecond, TotalUserTime.wMilliseconds);
}
*/
/*
 * recieve two FILETIME structures, calculate the difference and return the 
 * result in SYSTEMTIME structure
 */
SYSTEMTIME SubtractFileTimes(const FILETIME *after, const FILETIME *before)
{
	SYSTEMTIME result_systemtime;
		
	LARGE_INTEGER time_before, time_after, time_dif;
	FILETIME result_filetime;
	time_before.HighPart = before->dwHighDateTime;
	time_before.LowPart  = before->dwLowDateTime;
	time_after.HighPart  = after->dwHighDateTime;
	time_after.LowPart   = after->dwLowDateTime;

	time_dif.QuadPart = time_after.QuadPart-time_before.QuadPart;
	result_filetime.dwHighDateTime = time_dif.HighPart;
	result_filetime.dwLowDateTime = time_dif.LowPart;

	if (FileTimeToSystemTime (&result_filetime, &result_systemtime) == 0){
		fprintf(stderr, "FileTimeToSystemTime failed.\n");
		exit(1);
	}

	return result_systemtime;
}

/* 
 * Get SYSTEMTIME and number of iterations
 * Return the average time in microseconds
 */
long CalcAvgFileTime(const SYSTEMTIME *time, int iter)
{	
	DWORD total_mili = time->wMinute*60000+time->wSecond*1000+ time->wMilliseconds;

	return (total_mili * 1000)/iter;
}