/* QNX version */

/*
 * This test measures the time to manually generate a context switch of two 
 * processes in a loop by suspending the first and resuming the second and vice vera
 * in a loop.
 */

#define WAIT 999
#define MILLION 1000000
#define MAX_DATAPOINTS 50000
#define SLICE_FRAC .1
#define TRUE 1
#define FALSE 0

#include <stdio.h> 
#include <stdlib.h> 
#include <process.h> 
#include <sys/qnx_glob.h> 
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include "time_util.h"

/* global vars */
int g_pid0, g_pid1;
int g_iterations_per_slice;
/* Used to measure the average time of each slice */
struct timespec g_slice_datapoints[MAX_DATAPOINTS];

/* prototypes */
void sigchild_handler(int n);
void suspend_proc(int pid);
void resume_proc(int pid);
void fork_exec(char *filename, int instances);
void suspend_resume(int iterations, int warmup_iter);
void calc_stdev(int datapoints);

int  main(int argc, char** argv)
{
  int warmup_iter = 0;
  int iterations;
  int i;
 
  fprintf(stderr, "Compiled !\n");
  
  if ((argc != 3) && (argc != 4)){
    fprintf(stderr, "Usage: %s  <filename> <suspend-resume iterations> [warmup iterations]\n", argv[0]);
    exit(1);
  }
   iterations = atoi(argv[2]);
  if (argc == 4)
    warmup_iter = atoi(argv[3]);
 
  fork_exec(argv[1], 2);  //create two processes to suspend and resume 
 
  
  if (warmup_iter)
 	  suspend_resume(warmup_iter, TRUE);
  suspend_resume(iterations, FALSE);

  fprintf(stderr, "killing children.\n");
  
  kill (g_pid0, SIGCONT);
  kill (g_pid1, SIGCONT);
  
  if (kill (g_pid0, SIGKILL) < 0)
    perror("kill");
  if (kill (g_pid1, SIGKILL) < 0)
    perror ("kill");
  fprintf(stderr, "out of program.\n");
  return 0;
}

void fork_exec(char *filename, int instances)
{
  int i, pid;
  static char *arg_list[10];

  arg_list[0] = filename;
  arg_list[1] = NULL;
  for (i=0; i<instances; i++){
    pid = fork();
    if (pid < 0) {
      fprintf(stderr, "fork_exec: failed to fork");
      perror("fork_exec");
      exit(1);

    } else if (pid == 0) {
      if (execvp(filename, arg_list) < 0)
	perror("fork_exec");
      else 
	exit(0); /* just to make sure */
    } else { /* parent saves the PID */
      if (i == 0)
	g_pid0 = pid;
      else
	g_pid1 = pid;
    }
  }
}

void suspend_resume(int iter, int warmup)
{
  int i;
  struct timespec time_before, time_after, total_time;
  double total_microsec;
  int datapoints_count;
  int status;
 
  /* suspend one process */
  suspend_proc(g_pid1);
 
  if (warmup)
    fprintf(stderr, "warming up %d iter.\n", iter);
  else
    fprintf(stderr, "suspending & resuming %d iter.\n", iter);
  datapoints_count = 0;
  g_iterations_per_slice = (int)(iter *SLICE_FRAC + 0.5);
  if (g_iterations_per_slice == 0) {
    g_iterations_per_slice = 1000;  // dummy value for the warmup iterations 
  }
  clock_gettime(CLOCK_REALTIME, &time_before);
 
   
  for (i=0; i<iter; i++) { 
    if ((i % g_iterations_per_slice) == 0){
      clock_gettime(CLOCK_REALTIME, &g_slice_datapoints[datapoints_count]);
      datapoints_count++;
    }
  
    /* Suspend 0 and resume 1 */  
    kill(g_pid0, SIGSTOP);
    kill(g_pid1, SIGCONT);
  
    /* Suspend 1 and resume 0 */
    kill(g_pid1, SIGSTOP); 
    kill(g_pid0, SIGCONT);
  }
  clock_gettime(CLOCK_REALTIME, &time_after);
  clock_gettime(CLOCK_REALTIME, &g_slice_datapoints[datapoints_count]);
  datapoints_count++;

  if (warmup){ // just warming up, no need for calculations 
    return;
  }

  total_time = t_minus(time_after, time_before);
  total_microsec = total_time.tv_sec * MILLION + total_time.tv_nsec/1000; //total microseconds

  fprintf(stderr, "Average time(miscrosec); %5.3f; ", total_microsec / iter);
  fprintf(stderr, "Total time: %ld:%3.3f(Sec:MicroSec)\n", total_time.tv_sec, total_time.tv_nsec/1000.0);
  
  //calc_stdev(datapoints_count);
  fprintf(stderr, "Done.\n");
}



void suspend_proc(int pid)
{
  if (kill(pid, SIGSTOP) < 0) {
    perror ("suspend_proc");
    exit(1);
  } 
}

void resume_proc(int pid)
{
  if (kill(pid, SIGCONT) < 0) {
    perror ("resume_proc");
    exit(1);
  } 
}

void sigchild_handler(int n)
{
  int		 status;
 
   wait(&status); 
}


void calc_stdev(int datapoints)
{  
  static double slice_avg[MAX_DATAPOINTS];
  double diff, slices_avg, sum, stdev;
  struct timespec slice_total_time;  
  int i;
  double slices_avg_sum = 0.;

  for (i=1; i<datapoints; i++){
    slice_total_time = t_minus(g_slice_datapoints[i], g_slice_datapoints[i-1]);
    slice_avg[i-1] = (slice_total_time.tv_sec * MILLION + (slice_total_time.tv_nsec/1000.0)) / (g_iterations_per_slice*1.0);
    slices_avg_sum += slice_avg[i-1];
    /* fprintf(stderr, "val = %3.3f\n", slice_avg[i-1]); */
  }

  slices_avg = slices_avg_sum / ((datapoints-1) * 1.0);
  /* fprintf(stderr, "datapoints %d\n", datapoints);
     fprintf(stderr, "slices average is %3.3f\n", slices_avg);*/
  sum = 0;
  for (i=0; i<datapoints-1; i++){
    diff = slice_avg[i] - slices_avg;
    sum += diff*diff;
  }
  stdev = sqrt(sum / (datapoints - 1));
  stdev *= sqrt(g_iterations_per_slice);
  
  fprintf(stderr, "The STDDEV of %d slices is: %3.3f\n", datapoints-1, stdev);
}